@@ -2,7 +2,7 @@
  * LPRng - An Extended Print Spooler System
  *
  * Copyright 1988-1997, Patrick Powell, San Diego, CA
- *     papowell@sdsu.edu
+ *     papowell@astart.com
  * See LICENSE for conditions of use.
  *
  ***************************************************************************
@@ -11,15 +11,22 @@
  **************************************************************************/
 
 static char *const _id =
-"$Id: controlword.c,v 3.1 1996/12/28 21:40:11 papowell Exp $";
+"$Id: controlword.c,v 3.6 1997/12/24 20:10:12 papowell Exp $";
 
 #include "lp.h"
 #include "control.h"
 /**** ENDINCLUDE ****/
 
-#define PAIR(X) { #X, INTEGER_K, (void *)0, X }
+#undef PAIR
+#ifndef _UNPROTO_
+# define PAIR(X) { #X, INTEGER_K, (void *)0, X }
+#else
+# define __string(X) "X"
+# define PAIR(X) { __string(X), INTEGER_K, (void *)0, X }
+#endif
+
 static struct keywords controlwords[] = {
-PAIR(STATUS),
+PAIR(STATUs),
 PAIR(START),
 PAIR(STOP),
 PAIR(ENABLE),
@@ -42,6 +49,9 @@ PAIR(DEBUG),
 PAIR(HOLDALL),
 PAIR(NOHOLDALL),
 PAIR(CLAss),
+PAIR(DEFAULTQ),
+PAIR(ACTive),
+PAIR(REDO),
 {0}
 };
 
@@ -61,3 +71,14 @@ int Get_controlword( char *s )
 	}
 	return( 0 );
 }
+
+char *Get_controlstr( int c )
+{
+	int i;
+	for( i = 0; controlwords[i].keyword; ++i ){
+		if( controlwords[i].maxval == c ){
+			return( controlwords[i].keyword );
+		}
+	}
+	return( 0 );
+}
